package main

import (
	"fmt"
	"os"
	"strconv"
	"bufio"
	"strings"
	"log"
	"sort"
)

// A1, A2, B1, B2, B3, B4, B5, C.
const (  // iota is reset to 0
	A1 = iota
	A2 = iota
	B1 = iota
	B2 = iota
	B3 = iota
	B4 = iota
	B5 = iota
	C = iota
)

type LineParser func(*Licenses, []string)

type Quali struct {
	name string

	articlesCount int
	articlesPerTeacherCount float64
}

func (q *Quali) addArticle(a *Article) {
	q.articlesCount++;
	q.articlesPerTeacherCount += 1.0 / float64(len(a.authors))
}

func NewQuali(n string) Quali {
	q := Quali{
		name: n,
		articlesCount: 0,
		articlesPerTeacherCount: 0,
	}
	return q
}

type Teacher struct {
	name string
	code int
}

type QualiYear struct {
	q *Quali
	year int
}

type Vehicle struct {
	name string
	abbr string
	quali *Quali
	impact float64

	qualis []QualiYear
}

func (v *Vehicle) setYear(year int) {
	// search for the closest quali in the map
	var qy *QualiYear = nil
	for _, q := range v.qualis {
		if (qy == nil || year - qy.year >= year - q.year) && year >= q.year {
			qy = &q
		}
	}
	// fmt.Println("qy = ", qy.q.name)
	v.quali = qy.q
}

type Article struct {
	year int
	title string
	vehicle *Vehicle
	authors []*Teacher
}

func (a *Article) getQuali() *Quali{
	return a.vehicle.quali
}

type Licenses struct {
	teachers map[int]*Teacher
	vehicles map[string]*Vehicle
 	articles []*Article
 	qualis []Quali
}

func (l *Licenses) setYear(year int) {
	for _, v := range l.vehicles {
		v.setYear(year)
	}
	for _, a := range l.articles {
		a.vehicle.quali.addArticle(a)
	}
}

func findQualiId(qname string) int {
// 	fmt.Println("q = ", qname)
	switch qname {
		case "A1":
			return A1
		case "A2":
			return A2
		case "B1":
			return B1
		case "B2":
			return B2
		case "B3":
			return B3
		case "B4":
			return B4
		case "B5":
			return B5
		case "C":
			return C
		default:
			return -1
	}
}

func (l *Licenses) findQuali(qname string) *Quali {
	return &l.qualis[findQualiId(qname)]
}

// A1, A2, B1, B2, B3, B4, B5, C.
func NewQualis() []Quali {
	return []Quali{
		NewQuali("A1"),
		NewQuali("A2"),

		NewQuali("B1"),
		NewQuali("B2"),
		NewQuali("B3"),
		NewQuali("B4"),
		NewQuali("B5"),

		NewQuali("C"),
	}
}

func parseTeacher(l *Licenses, columns []string) {
	// <código>;<nome>;<data-nascimento>;<data-ingresso>;<coordenador?>

	// fmt.Println(columns)

	t := new(Teacher)
	code, _ := strconv.Atoi(columns[0])

	t.code = code
	t.name = columns[1]

	l.teachers[code] = t

	// fmt.Println("Here")
}

func parseVehicle(l *Licenses, columns []string) {
	// <sigla>;<nome>;<tipo>;<fator de impacto>;<issn>
	v := new(Vehicle)

	v.abbr = columns[0]
	v.name = columns[1]
	columns[3] = strings.Replace(columns[3], ",", ".", -1)
	impact, _ := strconv.ParseFloat(columns[3], 64)
	v.impact = impact
	v.qualis = nil

	l.vehicles[columns[0]] = v
}

func parseArticle(l *Licenses, columns []string) {
	// <ano>;<sigla veículo>;<título>;<lista de autores>;<número>;<volume periódico>;<local conferência>;<página inicial>;<página final>
	v := new(Article)

	year, _ := strconv.Atoi(columns[0])
	v.year = year
	v.vehicle = l.vehicles[columns[1]]
	v.title = columns[2]
	codes := strings.Split(columns[3], ",")
	
	// fmt.Println("authors", codes)

	for _, c := range codes {
		i, _ := strconv.Atoi(c)
		v.authors = append(v.authors, l.teachers[i])
	}

	l.articles = append(l.articles, v)
}

func parseQualis(l *Licenses, columns []string) {
	// <ano>;<sigla-veículo>;<Qualis>
// 	fmt.Println("abbr = ", columns[1])
	v := l.vehicles[columns[1]]
// 	fmt.Println("v = ", v)
	y, _ := strconv.Atoi(columns[0])

	qy := QualiYear{
		q: l.findQuali(columns[2]),
		year: y,
	}
	v.qualis = append(v.qualis, qy)
}

func readLines(l *Licenses, filename string, parser LineParser) {
	file, err := os.Open(filename)
	if err != nil {
		log.Fatal(err)
	}
	defer file.Close()
	
	scanner := bufio.NewScanner(file)
	scanner.Scan() // ignore first line
	for scanner.Scan() {
		s := strings.Split(scanner.Text(), ";")
		for i, c := range s {
			s[i] = strings.TrimSpace(c)
		}
		parser(l, s)
	}
	
	if err := scanner.Err(); err != nil {
		log.Fatal(err)
	}
}

type LineWriter func(*Licenses, *bufio.Writer)

func writeCSV(l *Licenses, filename string, header string, writer LineWriter) {
	fo, err := os.Create(filename)
	if err != nil {
		panic(err)
	}
	// close fo on exit and check for its returned error
	defer func() {
		if err := fo.Close(); err != nil {
			panic(err)
		}
	}()

	w := bufio.NewWriter(fo)

	_, err = fmt.Fprintln(w, header)
	if err != nil {
		panic(err)
	}

	writer(l, w)
	w.Flush()
}

func writeQualis(l *Licenses, w *bufio.Writer) {
	for _, q := range l.qualis {
		// fmt.Println(q)

		s := fmt.Sprintf("%.02f", q.articlesPerTeacherCount)
		s = strings.Replace(s, ".", ",", -1)

		_, err := fmt.Fprint(w, q.name, ";", q.articlesCount, ";", s, "\n")
		if err != nil {
			panic(err)
		}
	}
}

func writeArticles(l *Licenses, w *bufio.Writer) {
	sort.Slice(l.articles, func(i, j int) bool {
		a1 := l.articles[i]
		a2 := l.articles[j]

		// Este relatório deve ser ordenado por Qualis, em ordem decrescente (A1 primeiro), seguido por ano (decrescente) por sigla do veículo (crescente) e, por fim, pelo título da publicação (crescente).
		if a1.getQuali().name != a2.getQuali().name {
			return a1.getQuali().name < a2.getQuali().name
		}
		if a1.year != a2.year {
			return a1.year > a2.year
		}
		if a1.vehicle.abbr != a2.vehicle.abbr {
			return a1.vehicle.abbr < a2.vehicle.abbr
		}
		return a1.title < a2.title
	})
	for _, q := range l.articles {
		// fmt.Println(q)
		// <ano>;<sigla veículo>;<nome veículo>;<Qualis>;<fator de impacto>;<título da publicação>;<nomes dos docentes>

		var names []string = nil
		for _, a := range q.authors {
			names = append(names, a.name)
		}
		s := fmt.Sprint(strings.Join(names[:], ","))
		s2 := fmt.Sprintf("%.03f", q.vehicle.impact)
		s2 = strings.Replace(s2, ".", ",", -1)

		_, err := fmt.Fprint(w, q.year, ";", q.vehicle.abbr, ";", q.vehicle.name, ";", 
				q.vehicle.quali.name, ";", s2, ";", q.title, ";", s, "\n")
		if err != nil {
			panic(err)
		}
	}
}

func main() {
	args := os.Args[1:]

	// -d 01\in\docentes.csv -v 01\in\veiculos.csv -p 01\in\publicacoes.csv -q 01\in\qualis.csv -r 01\in\regras.csv -a 2017
	var teachers string
	var vehicles string
	var articles string
	var qualis string
	// var rules string // not needed

	var year int

	for i := 0; i < len(args); i += 2 {
		var arg string = args[i]
		var next string = args[i + 1]
		if arg == "-d" {
			teachers = next
		} else if arg == "-v" {
			vehicles = next
		} else if arg == "-p" {
			articles = next
		} else if arg == "-q" {
			qualis = next
		} else if arg == "-r" {
			// ignored
			fmt.Println("rule -r ", next, " will be ignored")
		} else if arg == "-a" {
			a, _ := strconv.Atoi(next)
			year = a
		}
	}

	fmt.Println("teachers ", teachers)
	fmt.Println("vehicles ", vehicles)
	fmt.Println("articles ", articles)
	fmt.Println("qualis ", qualis)

	fmt.Println("year ", year)

	l := Licenses{
		teachers: make(map[int]*Teacher),
		vehicles: make(map[string]*Vehicle),
		articles: nil,
		qualis: NewQualis(),
	}

	readLines(&l, teachers, parseTeacher)

	readLines(&l, vehicles, parseVehicle)
	readLines(&l, articles, parseArticle)
	readLines(&l, qualis, parseQualis)

	fmt.Println("teachers count", len(l.teachers))
	fmt.Println("vehicles count", len(l.vehicles))
	fmt.Println("articles count", len(l.articles))
	fmt.Println("qualis count", len(l.qualis))

	l.setYear(year)

	// Qualis;Qtd. Artigos;Média Artigos / Docente

	writeCSV(&l, "3-estatisticas.csv", "Qualis;Qtd. Artigos;Média Artigos / Docente", writeQualis)
	writeCSV(&l, "2-publicacoes.csv", "Ano;Sigla Veículo;Veículo;Qualis;Fator de Impacto;Título;Docentes", writeArticles)


	// fmt.Println(args)
}
